"""
pycraigslist.models.sessions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module handles all requests and construction of BeautifulSoup objects.
"""

import concurrent.futures
import tenacity
import requests
from bs4 import BeautifulSoup

HEADERS = {"headers": {"User-Agent": "Mozilla/5.0"}}
# Retry 10 times, starting with 0.01 second and doubling the delay every time
_RETRY_ARGS = {
    "wait": tenacity.wait.wait_random_exponential(multiplier=0.01, exp_base=2),
    "stop": tenacity.stop.stop_after_attempt(12),
}


class MaximumRequestsError(Exception):
    """Exceeds maximum get requests."""


def yield_html(url, **kwargs):
    """Yields HTML content(s) to caller."""
    try:
        # single request: a url string
        if isinstance(url, str):
            yield get_html(get_request(url, **parse_kwargs(kwargs)).content)
        # single request: a single url in a list or tuple
        elif isinstance(url, (list, tuple)) and len(url) == 1:
            yield get_html(get_request(url[0], **parse_kwargs(kwargs)).content)
        # multiple requests
        else:
            yield from map(
                get_html,
                (
                    response.content
                    for response in threaded_get_request(url, **parse_kwargs(kwargs))
                ),
            )
    except tenacity.RetryError:
        raise MaximumRequestsError("Maximum requests attempted - check network connection.")


def get_html(content):
    """Gets bs4.BeautifulSoup object from response content."""
    return BeautifulSoup(content, "html.parser")


@tenacity.retry(**_RETRY_ARGS)
def get_request(url, params=None):
    """Gets requests.models.Response object using requests.get.
    Retry request if request fails, with number of attempts and
    wait time specified in _RETRY_ARGS."""
    if params is None:
        params = HEADERS
    # don't add headers if params is {}
    elif params != {}:
        params.update(HEADERS)

    return requests.get(url, params=params, timeout=5)


def threaded_get_request(urls, **kwargs):
    """Yields requests from get_request concurrently."""
    yield from iter(
        concurrency(
            concurrent.futures.ThreadPoolExecutor, lambda args: get_request(*args), urls, **kwargs
        )
    )


def concurrency(PoolExecutor, map_func, *args, **kwargs):
    """General concurrency procedure to submit map-able
    functions to arguments."""
    # this procedure is designed to handle process pools and thread pools
    # zip args and kwarg values to make tuple of args
    zipped_args = zip(*args, *kwargs.values())
    with PoolExecutor(max_workers=5) as executor:
        futures = {
            # func must accept all positional arguments (kwargs assumed by args OK)
            executor.submit(map_func, arg_tuple)
            for arg_tuple in zipped_args
        }
        for future in concurrent.futures.as_completed(futures):
            yield future.result()


def parse_kwargs(kwargs):
    """Removes first kwarg value from list or tuple wrapper if wrapped and
    is the only value."""
    for key, value in kwargs.copy().items():
        if isinstance(value, (list, tuple)) and len(value) == 1:
            kwargs[key] = value[0]
    return kwargs



"""
pycraigslist.models.search_detail
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module gets and parses detailed posts from a Craigslist query.
"""

from . import sessions


def fetch_posts(posts_iter, **kwargs):
    """Fetches details from every post in posts_iter."""
    # yield all post values
    posts = tuple(posts_iter)
    posts_url = [post["url"] for post in posts]
    params = [{} for _ in posts_url]
    # get every post's HTML content
    post_htmls = sessions.yield_html(posts_url, params=params)
    kwargs["filters"] = invert_filters(kwargs["filters"])

    # yield detailed posts to caller
    for post_idx, html in enumerate(post_htmls):
        yield {**posts[post_idx], **get_post_details(html, **kwargs)}


def invert_filters(search_filters):
    """Inverts search filters for a faster look-up time."""
    # from {"cats_ok": {"url_key": "pets_cat", "value": 1, "attr": "cats are ok - purrr"},}
    # to {"cats are ok - purrr": {"cats_ok": "true"},}
    inv_filters = {}
    for key, value in search_filters.items():
        try:
            inv_filters[value["attr"]] = {key: "true"}
        except KeyError:
            if isinstance(value, dict):
                inv_filters.update({child_value: key for child_value in value})

    return inv_filters


def get_post_details(post_content, **kwargs):
    """Gets additional post details as dictionary."""
    addl_detail = {}
    addl_detail.update(get_geotag(post_content))
    addl_detail.update(get_address(post_content))
    addl_detail.update(get_post_attrs(post_content, **kwargs))
    include_body = kwargs.get("include_body")
    if include_body in (True, 1):
        addl_detail.update(get_body(post_content))

    return addl_detail


def get_geotag(post_content):
    """Gets geotag from post's HTML content."""
    post_map = post_content.find("div", {"id": "map"})
    geotag_attr = {"lat": "", "lon": ""}
    if post_map is not None:
        geotag_attr["lat"] = post_map.attrs["data-latitude"]
        geotag_attr["lon"] = post_map.attrs["data-longitude"]

    return geotag_attr


def get_address(post_content):
    """Gets address from post's HTML content."""
    post_address = post_content.find("div", {"class": "mapaddress"})
    address_attr = {"address": ""}
    if post_address is not None:
        address_attr["address"] = post_address.text

    return address_attr


def get_body(post_content):
    """Gets body from post's HTML content."""
    post_body = post_content.find("section", id="postingbody")
    body_attr = {"body": ""}
    if post_body is not None:
        # remove elements with null attributes
        post_body_text = (
            getattr(element, "text", element)
            for element in post_body
            if not getattr(element, "attrs", None)
        )
        # squelch post_body_text to str
        body_attr["body"] = str("".join(post_body_text).strip())

    return body_attr


def get_post_attrs(post_content, **kwargs):
    """Gets attributes from post's HTML content."""
    filters = kwargs["filters"]
    # attributes without a recognizable ID will be appended to "misc"
    post_attrs = {"misc": []}
    for attr in get_attrs(post_content):
        # add to post_attrs by reference
        parse_attrs(post_attrs, attr, filters)

    return post_attrs


def get_attrs(post_content):
    """Gets attribute keys from post's HTML content."""
    for attribute in post_content.find_all("p", {"class": "attrgroup"}):
        for attr in attribute.find_all("span"):
            attr_text = attr.text.strip()
            if attr_text:
                yield attr_text.lower()


def parse_attrs(post_attrs, attr, filters):
    """Parses attributes using search filters as reference. Adds parsed
    attributes to attributes collection (post_attrs) by reference."""
    if attr in filters:
        if isinstance(filters[attr], dict):
            # e.g. update with {"cats_ok": "true"} from {"cats are ok - purrr": {"cats_ok": "true"},}
            post_attrs.update(filters[attr])
        else:
            # update with attribute value
            post_attrs.update({filters[attr]: attr})
    elif ":" in attr:
        # some attr key, value are delimited by ':' - parse it to dict and update
        key_ = attr.split(":")[0].strip()
        attr_ = attr.split(":")[1].strip()
        post_attrs.update({key_: attr_})
    else:
        # attr not found in filter - append to list of anonymous attributes
        post_attrs["misc"].append(attr)



"""
pycraigslist.filters
~~~~~~~~~~~~~~~~~~~~

This module stores Craigslist search filters and categories.
"""


class search:
    """Store search filters by broad category."""

    base = {
        "query": {"url_key": "query", "value": None},
        "search_titles": {"url_key": "srchType", "value": "T"},
        "has_image": {"url_key": "hasPic", "value": 1},
        "posted_today": {"url_key": "postedToday", "value": 1},
        "bundle_duplicates": {"url_key": "bundleDuplicates", "value": 1},
        "search_distance": {"url_key": "search_distance", "value": None},
        "zip_code": {"url_key": "postal", "value": None},
    }

    community = {
        **base,
        **{
            # no unique filters for community
        },
    }

    events = {
        **base,
        **{
            # art/film
            "art": {"url_key": "event_art", "value": 1, "attr": "art/film"},
            "film": {"url_key": "event_art", "value": 1, "attr": "art/film"},
            # career
            "career": {"url_key": "event_career", "value": 1, "attr": "career"},
            # charitable
            "charitable": {"url_key": "event_fundraiser_vol", "value": 1, "attr": "charitable"},
            "fundraiser": {"url_key": "event_fundraiser_vol", "value": 1, "attr": "charitable"},
            # competiton
            "athletics": {"url_key": "event_athletics", "value": 1, "attr": "competition"},
            "competition": {"url_key": "event_athletics", "value": 1, "attr": "competition"},
            # dance
            "dance": {"url_key": "event_dance", "value": 1, "attr": "dance"},
            # fest/fair
            "festival": {"url_key": "event_festival", "value": 1, "attr": "fest/fair"},
            "fair": {"url_key": "event_festival", "value": 1, "attr": "fest/fair"},
            # fitness/health
            "fitness": {"url_key": "event_fitness_wellness", "value": 1, "attr": "fitness/health"},
            "health": {"url_key": "event_fitness_wellness", "value": 1, "attr": "fitness/health"},
            # food/drink
            "food": {"url_key": "event_food", "value": 1, "attr": "food/drink"},
            "drink": {"url_key": "event_food", "value": 1, "attr": "food/drink"},
            # free
            "free": {"url_key": "event_free", "value": 1, "attr": "free"},
            # kid friendly
            "kid_friendly": {"url_key": "event_kidfriendly", "value": 1, "attr": "kid friendly"},
            # literary
            "literary": {"url_key": "event_literary", "value": 1, "attr": "literary"},
            # music
            "music": {"url_key": "event_music", "value": 1, "attr": "music"},
            # outdoor
            "outdoor": {"url_key": "event_outdoor", "value": 1, "attr": "outdoor"},
            # sale
            "sale": {"url_key": "event_sale", "value": 1, "attr": "sale"},
            # singles
            "singles": {"url_key": "event_singles", "value": 1, "attr": "singles"},
            # tech
            "tech": {"url_key": "event_geek", "value": 1, "attr": "tech"},
        },
    }

    forsale = {
        **base,
        **{
            # price
            "min_price": {"url_key": "min_price", "value": None},
            "max_price": {"url_key": "max_price", "value": None},
            # make and model
            "make_model": {"url_key": "auto_make_model", "value": None},
            "min_year": {"url_key": "min_auto_year", "value": None},
            "max_year": {"url_key": "max_auto_year", "value": None},
            # odometer
            "min_miles": {"url_key": "min_auto_miles", "value": None},
            "max_miles": {"url_key": "max_auto_miles", "value": None},
            # engine displacement (cc)
            "min_engine_displacement": {"url_key": "min_engine_displacement_cc", "value": None},
            "max_engine_displacement": {"url_key": "max_engine_displacement_cc", "value": None},
        },
    }

    gigs = {
        **base,
        **{
            # paid/unpaid
            "is_paid": {"url_key": "is_paid", "value": 1},
        },
    }

    housing = {
        **base,
        **{
            # price
            "min_price": {"url_key": "min_price", "value": None},
            "max_price": {"url_key": "max_price", "value": None},
            # bedrooms
            "min_bedrooms": {"url_key": "min_bedrooms", "value": None},
            "max_bedrooms": {"url_key": "max_bedrooms", "value": None},
            # bathrooms
            "min_bathrooms": {"url_key": "min_bathrooms", "value": None},
            "max_bathrooms": {"url_key": "max_bathrooms", "value": None},
            # ft2
            "min_ft2": {"url_key": "minSqft", "value": None},
            "max_ft2": {"url_key": "maxSqft", "value": None},
            # private room
            "private_room": {"url_key": "private_room", "value": 1, "attr": "private room"},
            # private bath
            "private_bath": {"url_key": "private_bath", "value": 1, "attr": "private bath"},
            # cats ok
            "cats_ok": {"url_key": "pets_cat", "value": 1, "attr": "cats are ok - purrr"},
            # dogs ok
            "dogs_ok": {"url_key": "pets_dog", "value": 1, "attr": "dogs are ok - wooof"},
            # furnished
            "is_furnished": {"url_key": "is_furnished", "value": 1, "attr": "furnished"},
            # no smoking
            "no_smoking": {"url_key": "no_smoking", "value": 1, "attr": "no smoking"},
            # wheelchair access
            "wheelchair_acccess": {
                "url_key": "wheelchaccess",
                "value": 1,
                "attr": "wheelchair accessible",
            },
            # EV charging
            "ev_charging": {"url_key": "ev_charging", "value": 1, "attr": "ev charging"},
            # no appliation fee
            "no_application_fee": {"url_key": "application_fee", "value": 1},
            # no broker fee
            "no_broker_fee": {"url_key": "broker_fee", "value": 1},
        },
    }

    jobs = {
        **base,
        **{
            # internship
            "is_internship": {"url_key": "is_internship", "value": 1, "attr": "internship"},
            # non-profit
            "is_nonprofit": {
                "url_key": "is_nonprofit",
                "value": 1,
                "attr": "non-profit organization",
            },
            # telecommute
            "is_telecommuting": {
                "url_key": "is_telecommuting",
                "value": 1,
                "attr": "telecommuting okay",
            },
        },
    }

    resumes = {
        **base,
        **{
            # TODO: Please create an issue or PR if interested in this category.
        },
    }

    services = {
        **base,
        **{
            # no unique filters for services
        },
    }


class category:
    """Store niche category keys and descriptions within broad category."""

    @classmethod
    def get(cls, key):
        """Gets niche categories from broad category key."""
        index_table = {
            "community": cls.community,
            "events": cls.events,
            "forsale": cls.forsale,
            "gigs": cls.gigs,
            "housing": cls.housing,
            "jobs": cls.jobs,
            "resumes": cls.resumes,
            "services": cls.services,
        }
        return index_table[key]

    community = {
        "act": "activity partners",
        "ats": "artists",
        "kid": "childcare",
        "com": "general community",
        "grp": "groups",
        "vnn": "local news and views",
        "laf": "lost & found",
        "mis": "missed connections",
        "muc": "musicians",
        "pet": "pets",
        "pol": "politics",
        "rnr": "rants & raves",
        "rid": "rideshare",
        "vol": "volunteers",
    }

    events = {}

    forsale = {
        "ata": "antiques",
        "ppa": "appliances",
        "ppd": "appliances - by dealer",
        "app": "appliances - by owner",
        "ara": "arts & crafts",
        "ard": "arts & crafts - by dealer",
        "art": "arts & crafts - by owner",
        "sna": "atvs, utvs, snowmobiles",
        "snd": "atvs, utvs, snowmobiles - by dealer",
        "snw": "atvs, utvs, snowmobiles - by owner",
        "pta": "auto parts",
        "ptd": "auto parts - by dealer",
        "pts": "auto parts - by owner",
        "wta": "auto wheels & tires",
        "wtd": "auto wheels & tires - by dealer",
        "wto": "auto wheels & tires - by owner",
        "ava": "aviation",
        "avd": "aviation - by dealer",
        "avo": "aviation - by owner",
        "baa": "baby & kid stuff",
        "bad": "baby & kid stuff - by dealer",
        "bab": "baby & kid stuff - by owner",
        "bar": "barter",
        "bip": "bicycle parts",
        "bdp": "bicycle parts - by dealer",
        "bop": "bicycle parts - by owner",
        "bia": "bicycles",
        "bid": "bicycles - by dealer",
        "bik": "bicycles - by owner",
        "bpa": "boat parts & accessories",
        "bpd": "boat parts & accessories - by dealer",
        "bpo": "boat parts & accessories - by owner",
        "boo": "boats",
        "bod": "boats - by dealer",
        "boa": "boats - by owner",
        "bka": "books & magazines",
        "bkd": "books & magazines - by dealer",
        "bks": "books & magazines - by owner",
        "bfa": "business",
        "bfd": "business - by dealer",
        "bfs": "business - by owner",
        "cta": "cars & trucks",
        "ctd": "cars & trucks - by dealer",
        "cto": "cars & trucks - by owner",
        "ema": "cds / dvds / vhs",
        "emq": "cds / dvds / vhs - by dealer",
        "emd": "cds / dvds / vhs - by owner",
        "moa": "cell phones",
        "mod": "cell phones - by dealer",
        "mob": "cell phones - by owner",
        "cla": "clothing & accessories",
        "cld": "clothing & accessories - by dealer",
        "clo": "clothing & accessories - by owner",
        "cba": "collectibles",
        "cbd": "collectibles - by dealer",
        "clt": "collectibles - by owner",
        "syp": "computer parts",
        "sdp": "computer parts - by dealer",
        "sop": "computer parts - by owner",
        "sya": "computers",
        "syd": "computers - by dealer",
        "sys": "computers - by owner",
        "ela": "electronics",
        "eld": "electronics - by dealer",
        "ele": "electronics - by owner",
        "gra": "farm & garden",
        "zip": "free stuff",
        "fua": "furniture",
        "gms": "garage & moving sales",
        "foa": "general for sale",
        "fod": "general for sale - by dealer",
        "fur": "general for sale - by owner",
        "haa": "health and beauty",
        "had": "health and beauty - by dealer",
        "hab": "health and beauty - by owner",
        "hva": "heavy equipment",
        "hvd": "heavy equipment - by dealer",
        "hvo": "heavy equipment - by owner",
        "hsa": "household items",
        "hsd": "household items - by dealer",
        "hsh": "household items - by owner",
        "jwa": "jewelry",
        "jwd": "jewelry - by dealer",
        "jwl": "jewelry - by owner",
        "maa": "materials",
        "mad": "materials - by dealer",
        "mat": "materials - by owner",
        "mpa": "motorcycle parts & accessories",
        "mpd": "motorcycle parts & accessories - by dealer",
        "mpo": "motorcycle parts & accessories - by owner",
        "mca": "motorcycles/scooters",
        "mcd": "motorcycles/scooters - by dealer",
        "mcy": "motorcycles/scooters - by owner",
        "msa": "musical instruments",
        "msd": "musical instruments - by dealer",
        "msg": "musical instruments - by owner",
        "pha": "photo/video",
        "phd": "photo/video - by dealer",
        "pho": "photo/video - by owner",
        "rva": "recreational vehicles",
        "rvd": "recreational vehicles - by dealer",
        "rvs": "recreational vehicles - by owner",
        "sga": "sporting goods",
        "sgd": "sporting goods - by dealer",
        "spo": "sporting goods - by owner",
        "tia": "tickets",
        "tid": "tickets - by dealer",
        "tix": "tickets - by owner",
        "tla": "tools",
        "taa": "toys & games",
        "tad": "toys & games - by dealer",
        "tag": "toys & games - by owner",
        "tra": "trailers",
        "trb": "trailers - by dealer",
        "tro": "trailers - by owner",
        "vga": "video gaming",
        "vgd": "video gaming - by dealer",
        "vgm": "video gaming - by owner",
        "waa": "wanted",
        "wad": "wanted - by dealer",
        "wan": "wanted - by owner",
    }

    gigs = {
        "cpg": "computer gigs",
        "crg": "creative gigs",
        "cwg": "crew gigs",
        "dmg": "domestic gigs",
        "evg": "event gigs",
        "lbg": "labor gigs",
        "tlg": "talent gigs",
        "wrg": "writing gigs",
    }

    housing = {
        "apa": "apartments / housing for rent",
        "swp": "housing swap",
        "off": "office & commercial",
        "prk": "parking & storage",
        "rea": "real estate",
        "reb": "real estate - by dealer",
        "reo": "real estate - by owner",
        "roo": "rooms & shares",
        "sub": "sublets & temporary",
        "vac": "vacation rentals",
        "hou": "wanted: apts",
        "rew": "wanted: real estate",
        "sha": "wanted: room/share",
        "sbw": "wanted: sublet/temp",
    }

    jobs = {
        "acc": "accounting/finance",
        "ofc": "admin/office",
        "egr": "architect/engineer/cad",
        "med": "art/media/design",
        "bus": "business/mgmt",
        "csr": "customer service",
        "edu": "education/teaching",
        "etc": "et cetera",
        "fbh": "food/beverage/hospitality",
        "lab": "general labor",
        "gov": "government",
        "hea": "healthcare",
        "hum": "human resource",
        "lgl": "legal/paralegal",
        "mnu": "manufacturing",
        "mar": "marketing/advertising/pr",
        "npo": "nonprofit",
        "rej": "real estate",
        "ret": "retail/wholesale",
        "sls": "sales",
        "spa": "salon/spa/fitness",
        "sci": "science/biotech",
        "sec": "security",
        "trd": "skilled trades/artisan",
        "sof": "software/qa/dba/etc",
        "sad": "systems/networking",
        "tch": "technical support",
        "trp": "transportation",
        "tfr": "tv/film/video/radio",
        "web": "web/html/info design",
        "wri": "writing/editing",
    }

    resumes = {}

    services = {
        "aos": "automotive services",
        "bts": "beauty services",
        "cms": "cell phone / mobile services",
        "cps": "computer services",
        "crs": "creative services",
        "cys": "cycle services",
        "evs": "event services",
        "fgs": "farm & garden services",
        "fns": "financial services",
        "hss": "household services",
        "lbs": "labor / hauling / moving",
        "lgs": "legal services",
        "lss": "lessons & tutoring",
        "mas": "marine services",
        "pas": "pet services",
        "rts": "real estate services",
        "sks": "skilled trade services",
        "biz": "small biz ads",
        "trv": "travel/vacation services",
        "wet": "writing / editing / translation",
    }



"""
pycraigslist.tests.specs.specs
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module stores static information for testing.
"""


class filters:
    """Build suite of filters used in searches by category."""

    housing = (
        {"min_bedrooms": 1, "max_price": 2000, "cats_ok": 1},
        {"max_price": 1000, "dogs_ok": True},
    )



"""
pycraigslist.models.filters
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module parses and gets additional Craigslist query filters.
"""

from . import sessions


def get_addl_readable(url):
    """Gets additional Craigslist query filters in readable format."""
    filters = get_addl(url)
    for key, value in filters.copy().items():
        if isinstance(value, dict):
            # get only filter names
            # e.g. ['apartment', 'condo'] from {'apartment': '1', 'condo': '2'}
            filters[key] = list(value)

    return filters


def get_addl(url):
    """Gets additional Craigslist query filters."""
    # additional filters constitute niche filters for a category
    # e.g. 'rent_period' is an additional filter for apartments / housing for rent
    filters = {}
    search_html = next(sessions.yield_html(url))

    for filter_item in search_html.find_all("div", class_="search-attribute"):
        filter_key = filter_item.attrs["data-attr"]
        filter_labels = filter_item.find_all("label")
        filters[filter_key] = {
            opt.text.strip(): opt.find("input").get("value") for opt in filter_labels
        }

    return filters


def parse(filters, search_filters):
    """Parses and validates url and filters using categorical search filters as
    template reference. Returns filters for requests.get params."""
    # iterate over a copy of filters
    for key, value in filters.copy().items():
        if key in ("searchNearby", "s"):
            # skip default filters found in all queries
            continue
        try:
            # substitute and remove key for url_key
            url_key = search_filters[key]["url_key"]
            filters[url_key] = parse_value(value)
            if url_key != key:
                # delete old key
                del filters[key]
        except KeyError:
            # some filters allow multiple values - assign all specified by user
            try:
                filters[key] = [
                    search_filters[key][parsed_val] for parsed_val in parse_value(value)
                ]
            except (KeyError, TypeError):
                raise ValueError("filter '%s' is or has a bad value" % key)

    return filters


def parse_value(filter_value):
    """Validates and further parses search filter values."""
    # returns filter value(s) as list
    if isinstance(filter_value, (float, str)):
        return [filter_value]
    # returns bool or int as int
    if isinstance(filter_value, (bool, int)):
        return [int(filter_value)]
    # recursively parses values if filter_value is an iterable
    return iter(parse_value(value)[0] for value in filter_value)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pycraigslist', 'pycraigslist.models']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.25.0', 'beautifulsoup4>=4.9.0', 'tenacity>=7.0.0']

setup(name='pycraigslist',
      version='0.4.0',
      description='pycraigslist API library',
      author='Ira Horecka',
      author_email='ira89@icloud.com',
      url='https://github.com/irahorecka/pycraigslist',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.5',
     )



"""
pycraigslist.base
~~~~~~~~~~~~~~~~~

This module handles all communication from the user to the Craigslist API.
"""

from . import models
from .utils import parse_limit


class BaseAPI:
    """Base class for interfacing with the Craigslist API."""

    category = ""
    search_filters = {}

    def __init__(self, site="sfbay", area="", filters=None, **kwargs):
        self.site = site
        self.area = area
        self.search_filters = {**self.search_filters, **models.filters.get_addl(self.url)}
        self.filters = {"searchNearby": 1, "s": 0}
        # **kwargs will override filters if matching key exists
        if isinstance(filters, dict):
            self.filters.update(filters)
        self.filters.update(**kwargs)
        # parse self.filters to construct valid HTTP parameters
        self.filters = models.filters.parse(self.filters, self.search_filters)

    @parse_limit
    def search(self, limit=None):
        """Yields Craigslist posts as dictionary."""
        yield from models.search.fetch_posts(
            self.url, self.filters, category=self.category, limit=limit
        )

    @parse_limit
    def search_detail(self, limit=None, include_body=False):
        """Yields detailed Craigslist posts as dictionary."""
        yield from models.search_detail.fetch_posts(
            self.search(limit=limit), filters=self.search_filters, include_body=include_body
        )

    @property
    def count(self):
        """Gets approximate number of Craigslist posts."""
        return models.search.get_total_post_count(self.url, self.filters)

    @property
    def url(self):
        """Builds Craigslist query URL."""
        if not self.area:
            return "https://%s.craigslist.org/search/%s" % (self.site, self.category)
        return "https://%s.craigslist.org/search/%s/%s" % (self.site, self.area, self.category)

    @classmethod
    def get_filters(cls):
        """Gets readable Craigslist query filters for given class.
        The intention is to provide user with valid filter keys and values,
        therefore the returned dictionary is catered for ease of interpretation."""
        category = cls.__name__ if not cls.category else cls.category
        # set sfbay classified as default craigslist search url
        category_url = "https://sfbay.craigslist.org/search/%s" % category
        readable_filters = {
            key: "..." if value["value"] is None else "True/False"
            for key, value in cls.search_filters.items()
        }
        readable_filters.update(models.filters.get_addl_readable(category_url))

        return readable_filters


class ParentMethods:
    """Provides parent classes with additional class methods."""

    @classmethod
    def get_categories(cls):
        """Gets valid Craigslist categories of instance."""
        from . import filters

        return filters.category.get(cls.__name__)
